import pytest

import cryojax.simulator as cxs
from cryojax.constants import b_factor_to_variance
from cryojax.io import read_array_from_mrc, read_atoms_from_pdb
from cryojax.simulator import DiscreteStructuralEnsemble


@pytest.fixture
def voxel_volume(sample_mrc_path):
    real_voxel_grid = read_array_from_mrc(sample_mrc_path)
    return (
        cxs.FourierVoxelGridVolume.from_real_voxel_grid(real_voxel_grid, pad_scale=1.3),
    )


@pytest.fixture
def gmm_volume(sample_pdb_path):
    atom_positions, atom_types, b_factors = read_atoms_from_pdb(
        sample_pdb_path,
        center=True,
        selection_string="not element H",
        loads_b_factors=True,
    )
    scattering_factor_parameters = cxs.PengScatteringFactorParameters(atom_types)
    return cxs.GaussianMixtureVolume(
        positions=atom_positions,
        amplitudes=scattering_factor_parameters.a,
        variances=b_factor_to_variance(
            scattering_factor_parameters.b + b_factors[:, None]
        ),
    )


@pytest.mark.parametrize(
    "volume",
    [("voxel_volume"), ("gmm_volume")],
)
def test_conformation(volume, request):
    volume = request.getfixturevalue(volume)
    conformational_space = tuple([volume for _ in range(3)])
    volume = DiscreteStructuralEnsemble(conformational_space, conformation=0)
    _ = volume.compute_volume_representation()



from .base_parametrisation import (
    AbstractEnsembleParametrisation as AbstractEnsembleParametrisation,
    AbstractPotentialParametrisation as AbstractPotentialParametrisation,
    AbstractVolumeParametrisation as AbstractVolumeParametrisation,
    AbstractVolumeRepresentation as AbstractVolumeRepresentation,
)
from .ensemble import DiscreteStructuralEnsemble as DiscreteStructuralEnsemble
from .potential import (
    AbstractPengPotential as AbstractPengPotential,
    AbstractTabulatedPotential as AbstractTabulatedPotential,
    PengAtomPotential as PengAtomPotential,
    PengScatteringFactorParameters as PengScatteringFactorParameters,
)
from .representations import (
    AbstractAtomVolume as AbstractAtomVolume,
    AbstractPointCloudVolume as AbstractPointCloudVolume,
    AbstractVoxelVolume as AbstractVoxelVolume,
    FourierVoxelGridVolume as FourierVoxelGridVolume,
    FourierVoxelSplineVolume as FourierVoxelSplineVolume,
    GaussianMixtureVolume as GaussianMixtureVolume,
    RealVoxelGridVolume as RealVoxelGridVolume,
)



import pytest

import cryojax.simulator as cxs


def test_deprecated():
    # Old CTF aliases
    with pytest.warns(DeprecationWarning):
        obj = cxs.AberratedAstigmaticCTF
        assert obj is cxs.AstigmaticCTF

    with pytest.warns(DeprecationWarning):
        obj = cxs.CTF
        assert obj is cxs.AstigmaticCTF



"""
Methods for integrating the volume directly onto the exit plane.
"""

from abc import abstractmethod
from typing import Generic, TypeVar
from typing_extensions import override

import equinox as eqx
from equinox import AbstractClassVar
from jaxtyping import Array, Complex, Float

from ..._image_config import AbstractImageConfig
from ..._volume_parametrisation import AbstractVoxelVolume
from ..base_integrator import AbstractVolumeIntegrator, VolRepT


VoxelVolT = TypeVar("VoxelVolT", bound="AbstractVoxelVolume")


class AbstractDirectIntegrator(
    AbstractVolumeIntegrator[VolRepT], Generic[VolRepT], strict=True
):
    """Base class for a method of integrating a volume onto
    the exit plane.
    """

    is_projection_approximation: AbstractClassVar[bool]

    @abstractmethod
    @override
    def integrate(
        self,
        volume_representation: VolRepT,
        image_config: AbstractImageConfig,
        outputs_real_space: bool = False,
    ) -> (
        Complex[
            Array,
            "{image_config.padded_y_dim} {image_config.padded_x_dim//2+1}",
        ]
        | Complex[Array, "{image_config.padded_y_dim} {image_config.padded_x_dim}"]
        | Float[Array, "{image_config.padded_y_dim} {image_config.padded_x_dim}"]
    ):
        raise NotImplementedError


class AbstractDirectVoxelIntegrator(
    AbstractDirectIntegrator[VoxelVolT], Generic[VoxelVolT], strict=True
):
    outputs_integral: eqx.AbstractVar[bool]



from typing import Optional
from typing_extensions import override

import jax
import jax.numpy as jnp
from jaxtyping import Array, Int, PRNGKeyArray

from ....jax_util import error_if_negative
from ..base_parametrisation import (
    AbstractEnsembleParametrisation,
    AbstractVolumeRepresentation,
)


class DiscreteStructuralEnsemble(AbstractEnsembleParametrisation, strict=True):
    """Abstraction of an ensemble with discrete conformational
    heterogeneity.
    """

    conformational_space: tuple[AbstractVolumeRepresentation, ...]
    conformation: Int[Array, ""]

    def __init__(
        self,
        conformational_space: tuple[AbstractVolumeRepresentation, ...],
        conformation: int | Int[Array, ""],
    ):
        """**Arguments:**

        - `conformational_space`: A tuple of `AbstractStructureRepresentation`s.
        - `conformation`: A conformation with a discrete index at which to evaluate
                          the tuple.
        """
        self.conformational_space = conformational_space
        self.conformation = jnp.asarray(error_if_negative(conformation), dtype=int)

    @override
    def compute_volume_representation(
        self, rng_key: Optional[PRNGKeyArray] = None
    ) -> AbstractVolumeRepresentation:
        """Map to the volume at `conformation`.

        **Arguments:**

        - `rng_key`:
            Not used in this implementation, but optionally
            included for other implementations.
        """
        funcs = [
            lambda i=i: self.conformational_space[i]
            for i in range(len(self.conformational_space))
        ]
        volume = jax.lax.switch(self.conformation, funcs)

        return volume



import abc
from typing import Any, Generic, TypeVar

import equinox as eqx
from jaxtyping import Array

from .._image_config import AbstractImageConfig
from .._volume_parametrisation import AbstractVolumeRepresentation


VolRepT = TypeVar("VolRepT", bound="AbstractVolumeRepresentation")


class AbstractVolumeIntegrator(eqx.Module, Generic[VolRepT], strict=True):
    """Base class for a method of integrating a volume into an image."""

    @abc.abstractmethod
    def integrate(
        self,
        volume_representation: VolRepT,
        image_config: AbstractImageConfig,
        *args: Any,
        **kwargs: Any,
    ) -> Array:
        raise NotImplementedError



from .tabulated_potential import (
    AbstractPengPotential as AbstractPengPotential,
    AbstractTabulatedPotential as AbstractTabulatedPotential,
    PengAtomPotential as PengAtomPotential,
    PengScatteringFactorParameters as PengScatteringFactorParameters,
)
